==> App-bif-0.1.0_23.tar.gz: App-bif-0.1.0_23/lib/App/bif/export.pm <==
package App::bif::export;
use strict;
use warnings;
use App::bif::Context;
use AnyEvent;
use Bif::Client;
use Coro;

our $VERSION = '0.1.0_23';

sub run {
    my $ctx = shift;
    $ctx->{no_pager}++;    # causes problems with something in Coro?
    $ctx = App::bif::Context->new($ctx);

    # Consider upping PRAGMA cache_size? Or handle that in Bif::Role::Sync?
    my $db = $ctx->dbw;

    my @pinfo;
    foreach my $path ( @{ $ctx->{path} } ) {
        my $pinfo = $ctx->get_project($path);
        push( @pinfo, $pinfo );
    }

    my @locations = $db->get_hub_repos( $ctx->{hub} );
    $ctx->err( 'HubNotFound', 'hub not found: %s', $ctx->{hub} )
      unless @locations;

    my $hub = $locations[0];

    my @new_pinfo;
    foreach my $pinfo (@pinfo) {
        my $exists =
          eval { $ctx->get_project( $pinfo->{path}, $hub->{name} ) };

        if ($exists) {
            if ( $exists->{uuid} eq $pinfo->{uuid} ) {
                print "Already exported to $hub->{name}: $pinfo->{path}\n";
                next;
            }
            else {
                return $ctx->err( 'PathExists',
                    'path exists at destination: %s',
                    $pinfo->{path} );
            }
        }
        push( @new_pinfo, $pinfo );
    }

    return $ctx->ok('Export') unless @new_pinfo;
    @pinfo = @new_pinfo;

    $|++;    # no buffering
    my $error;
    my $cv = AE::cv;

    my $client = Bif::Client->new(
        db       => $db,
        location => $hub->{location},
        debug    => $ctx->{debug},
        debug_bs => $ctx->{debug_bs},
        on_error => sub {
            $error = shift;
            $cv->send;
        },
    );

    my $stderr = $client->child->stderr;

    my $stderr_watcher;
    $stderr_watcher = AE::io $stderr, 0, sub {
        my $line = $stderr->getline;
        if ( !defined $line ) {
            undef $stderr_watcher;
            return;
        }
        print STDERR "$hub->{name}: $line";
    };

    my $coro = async {
        eval {
            $db->txn(
                sub {
                    $ctx->update_repo(
                        {
                            related_update_id => $ctx->{update_id},
                            message =>
                              "export @{$ctx->{path}} $hub->{location}",
                        }
                    );

                    foreach my $pinfo (@pinfo) {
                        my $uid = $db->nextval('updates');
                        $db->xdo(
                            insert_into => 'updates',
                            values      => {
                                id        => $uid,
                                parent_id => $hub->{first_update_id},
                                author    => $ctx->{user}->{name},
                                email     => $ctx->{user}->{email},
                                message   => "Imported $pinfo->{path}",
                            },
                        );

                        $db->xdo(
                            insert_into => 'hub_deltas',
                            values      => {
                                update_id  => $uid,
                                hub_id     => $hub->{id},
                                project_id => $pinfo->{id},
                            },
                        );

                        $db->xdo(
                            delete_from => 'hub_related_projects',
                            where       => {
                                hub_id     => $db->get_local_hub_id,
                                project_id => $pinfo->{id},
                            },
                        );

                        my $msg = "[Exported to $hub->{location}]";
                        if ( $ctx->{message} ) {
                            $msg .= "\n\n$ctx->{message}\n";
                        }

                        $ctx->{update_id} = $db->nextval('updates');

                        $db->xdo(
                            insert_into => 'updates',
                            values      => {
                                id        => $ctx->{update_id},
                                parent_id => $pinfo->{first_update_id},
                                author    => $ctx->{user}->{name},
                                email     => $ctx->{user}->{email},
                                message   => $msg,
                            },
                        );

                        $db->xdo(
                            insert_into => 'func_update_project',
                            values      => {
                                id        => $pinfo->{id},
                                update_id => $ctx->{update_id},
                                hub_uuid  => $hub->{uuid},
                            },
                        );

                        $db->xdo(
                            insert_into => 'func_merge_updates',
                            values      => { merge => 1 },
                        );

                        $client->on_update(
                            sub {
                                $ctx->lprint(
                                    "$hub->{name} [$pinfo->{path}]: $_[0]");
                            }
                        );

                        my $status = $client->sync_hub( $hub->{id} );
                        print "\n";

                        unless ( $status eq 'RepoSync' ) {
                            $db->rollback;
                            $error = "unexpected status received: $status";
                            last;
                        }

                        $status = $client->sync_project( $pinfo->{id} );
                        print "\n";

                        unless ( $status eq 'ProjectSync'
                            or $status eq 'ProjectMatch' )
                        {
                            $db->rollback;
                            $error = "unexpected status received: $status";
                            last;
                        }

                    }

                    # Catch up on errors
                    undef $stderr_watcher;
                    $stderr->blocking(0);
                    while ( my $line = $stderr->getline ) {
                        print STDERR "$hub->{name}: $line";
                    }

                    return;
                }
            );
        };

        if ($@) {
            $error .= $@;
            print "\n";
        }

        $client->disconnect;
        return $cv->send( !$error );
    };

    if ( $cv->recv ) {
        print "Project(s) exported: @{ $ctx->{path} }\n";
        return $ctx->ok('Export');
    }
    return $ctx->err( 'Unknown', $error );
}

1;
__END__

=head1 NAME

bif-export -  export a project to a remote hub

=head1 VERSION

0.1.0_23 (2014-06-04)

=head1 SYNOPSIS

    bif export PATH... HUB [OPTIONS...]

=head1 DESCRIPTION

Export a project to a hub.

=head1 ARGUMENTS & OPTIONS

=over

=item PATH

The path of the local project to export.  An error will be raised if a
project with the same path exists at the remote HUB.

=item HUB

The name of a previously registered hub.

=item --message, -m MESSAGE

Add the optional MESSAGE to the update created for this action.

=back

=head1 SEE ALSO

L<bif>(1)

=head1 AUTHOR

Mark Lawrence E<lt>nomad@null.netE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2013-2014 Mark Lawrence <nomad@null.net>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

==> App-bif-0.1.0_24.tar.gz: App-bif-0.1.0_24/lib/App/bif/export.pm <==
package App::bif::export;
use strict;
use warnings;
use App::bif::Context;
use AnyEvent;
use Bif::Client;
use Coro;

our $VERSION = '0.1.0_24';

my $stderr;
my $stderr_watcher;

sub cleanup_errors {
    my $hub = shift;

    undef $stderr_watcher;
    $stderr->blocking(0);

    while ( my $line = $stderr->getline ) {
        print STDERR "$hub: $line";
    }

    return;
}

sub run {
    my $ctx = shift;
    $ctx->{no_pager}++;    # causes problems with something in Coro?
    $ctx = App::bif::Context->new($ctx);

    # Consider upping PRAGMA cache_size? Or handle that in Bif::Role::Sync?
    my $db = $ctx->dbw;

    my @pinfo;
    foreach my $path ( @{ $ctx->{path} } ) {
        my $pinfo = $ctx->get_project($path);
        push( @pinfo, $pinfo );
    }

    my @locations = $db->get_hub_repos( $ctx->{hub} );
    $ctx->err( 'HubNotFound', 'hub not found: %s', $ctx->{hub} )
      unless @locations;

    my $hub = $locations[0];

    my @new_pinfo;
    foreach my $pinfo (@pinfo) {
        my $exists =
          eval { $ctx->get_project( $pinfo->{path}, $hub->{name} ) };

        if ($exists) {
            if ( $exists->{uuid} eq $pinfo->{uuid} ) {
                print "Already exported to $hub->{name}: $pinfo->{path}\n";
                next;
            }
            else {
                return $ctx->err( 'PathExists',
                    'path exists at destination: %s',
                    $pinfo->{path} );
            }
        }
        push( @new_pinfo, $pinfo );
    }

    return $ctx->ok('Export') unless @new_pinfo;
    @pinfo = @new_pinfo;

    $|++;    # no buffering
    my $error;
    my $cv = AE::cv;

    my $client = Bif::Client->new(
        db       => $db,
        location => $hub->{location},
        debug    => $ctx->{debug},
        debug_bs => $ctx->{debug_bs},
        on_error => sub {
            $error = shift;
            $cv->send;
        },
    );

    $stderr = $client->child->stderr;

    $stderr_watcher = AE::io $stderr, 0, sub {
        my $line = $stderr->getline;
        if ( !defined $line ) {
            undef $stderr_watcher;
            return;
        }
        print STDERR "$hub->{name}: $line";
    };

    my $coro = async {
        eval {
            $db->txn(
                sub {
                    $ctx->update_repo(
                        {
                            related_update_id => $ctx->{update_id},
                            message =>
                              "export @{$ctx->{path}} $hub->{location}",
                        }
                    );

                    foreach my $pinfo (@pinfo) {
                        my $uid = $db->nextval('updates');
                        $db->xdo(
                            insert_into => 'updates',
                            values      => {
                                id        => $uid,
                                parent_id => $hub->{first_update_id},
                                author    => $ctx->{user}->{name},
                                email     => $ctx->{user}->{email},
                                message   => "Imported $pinfo->{path}",
                            },
                        );

                        $db->xdo(
                            insert_into => 'hub_deltas',
                            values      => {
                                update_id  => $uid,
                                hub_id     => $hub->{id},
                                project_id => $pinfo->{id},
                            },
                        );

                        $db->xdo(
                            delete_from => 'hub_related_projects',
                            where       => {
                                hub_id     => $db->get_local_hub_id,
                                project_id => $pinfo->{id},
                            },
                        );

                        my $msg = "[Exported to $hub->{location}]";
                        if ( $ctx->{message} ) {
                            $msg .= "\n\n$ctx->{message}\n";
                        }

                        $ctx->{update_id} = $db->nextval('updates');

                        $db->xdo(
                            insert_into => 'updates',
                            values      => {
                                id        => $ctx->{update_id},
                                parent_id => $pinfo->{first_update_id},
                                author    => $ctx->{user}->{name},
                                email     => $ctx->{user}->{email},
                                message   => $msg,
                            },
                        );

                        $db->xdo(
                            insert_into => 'func_update_project',
                            values      => {
                                id        => $pinfo->{id},
                                update_id => $ctx->{update_id},
                                hub_uuid  => $hub->{uuid},
                            },
                        );

                        $db->xdo(
                            insert_into => 'func_merge_updates',
                            values      => { merge => 1 },
                        );
                    }

                    $client->on_update(
                        sub {
                            $ctx->lprint("$hub->{name}         : $_[0]");
                        }
                    );

                    my $status = $client->sync_hub( $hub->{id} );

                    if ( $status ne 'RepoSync' ) {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );
                    }

                    $status = $client->transfer_hub_updates;
                    if ( $status ne 'TransferHubUpdates' ) {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );
                    }
                    print "\n";

                    $status = $client->sync_projects;

                    unless ( $status eq 'ProjectSync' ) {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );
                    }

                    $client->on_update(
                        sub {
                            $ctx->lprint("$hub->{name} (topics): $_[0]");
                        }
                    );
                    $status = $client->transfer_project_related_updates;

                    if ( $status ne 'TransferProjectRelatedUpdates' ) {
                        $db->rollback;
                        $error = "unexpected status received: $status";
                        return cleanup_errors( $hub->{name} );
                    }
                    print "\n";

                    return;
                }
            );
        };

        if ($@) {
            $error .= $@;
            print "\n";
        }

        $client->disconnect;
        return $cv->send( !$error );
    };

    if ( $cv->recv ) {
        print "Project(s) exported: @{ $ctx->{path} }\n";
        return $ctx->ok('Export');
    }
    return $ctx->err( 'Unknown', $error );
}

1;
__END__

=head1 NAME

bif-export -  export a project to a remote hub

=head1 VERSION

0.1.0_24 (2014-06-13)

=head1 SYNOPSIS

    bif export PATH... HUB [OPTIONS...]

=head1 DESCRIPTION

Export a project to a hub.

=head1 ARGUMENTS & OPTIONS

=over

=item PATH

The path of the local project to export.  An error will be raised if a
project with the same path exists at the remote HUB.

=item HUB

The name of a previously registered hub.

=item --message, -m MESSAGE

Add the optional MESSAGE to the update created for this action.

=back

=head1 SEE ALSO

L<bif>(1)

=head1 AUTHOR

Mark Lawrence E<lt>nomad@null.netE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2013-2014 Mark Lawrence <nomad@null.net>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.
